"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// @flow
const fs = require("fs");
const graphile_build_1 = require("graphile-build");
exports.SchemaBuilder = graphile_build_1.SchemaBuilder;
const graphile_build_pg_1 = require("graphile-build-pg");
exports.formatSQLForDebugging = graphile_build_pg_1.formatSQLForDebugging;
function errorProxy(message) {
    return new Proxy({}, {
        get: () => {
            throw new Error(message);
        },
    });
}
const ensureValidPlugins = (name, arr) => {
    if (!Array.isArray(arr)) {
        throw new Error(`Option '${name}' should be an array`);
    }
    for (let i = 0, l = arr.length; i < l; i++) {
        const fn = arr[i];
        if (typeof fn !== "function") {
            throw new Error(`Option '${name}' should be an array of functions, found '${fn == null ? String(fn) : typeof fn}' at index ${i}`);
        }
    }
};
/*
 * BELOW HERE IS DEPRECATED!!
 */
exports.inflections = errorProxy("`inflections` was deprecated in v4.0.0-beta.7; instead please write an inflector plugin: https://www.graphile.org/postgraphile/inflection/");
exports.postGraphileBaseOverrides = errorProxy("Passing an inflector via PostGraphile options was deprecated in v4.0.0-beta.7; instead please write an inflector plugin: https://www.graphile.org/postgraphile/inflection/");
exports.postGraphileClassicIdsOverrides = errorProxy("Passing an inflector via PostGraphile options was deprecated in v4.0.0-beta.7; instead please write an inflector plugin: https://www.graphile.org/postgraphile/inflection/");
exports.postGraphileInflection = errorProxy("Passing an inflector via PostGraphile options was deprecated in v4.0.0-beta.7; instead please write an inflector plugin: https://www.graphile.org/postgraphile/inflection/");
exports.postGraphileClassicIdsInflection = errorProxy("Passing an inflector via PostGraphile options was deprecated in v4.0.0-beta.7; instead please write an inflector plugin: https://www.graphile.org/postgraphile/inflection/");
/*
 * ABOVE HERE IS DEPRECATED.
 */
exports.PostGraphileInflectionPlugin = function (builder) {
    builder.hook("inflection", (inflection) => {
        const previous = inflection.enumName;
        // Overwrite directly so that we don't lose the 'extend' hints
        Object.assign(inflection, {
            enumName(value) {
                return this.constantCase(previous.call(this, value));
            },
        });
        return inflection;
    });
};
exports.PostGraphileClassicIdsInflectionPlugin = function (builder) {
    builder.hook("inflection", (inflection) => {
        const previous = inflection._columnName;
        // Overwrite directly so that we don't lose the 'extend' hints
        Object.assign(inflection, {
            _columnName(attr, options) {
                const previousValue = previous.call(this, attr, options);
                return (options && options.skipRowId) || previousValue !== "id"
                    ? previousValue
                    : this.camelCase("rowId");
            },
        });
        return inflection;
    });
};
const awaitKeys = async (obj) => {
    const result = {};
    for (const k in obj) {
        if (Object.prototype.hasOwnProperty.call(obj, k)) {
            result[k] = await obj[k];
        }
    }
    return result;
};
const getPostGraphileBuilder = async (pgConfig, schemas, options = {}) => {
    // @ts-ignore
    if (options.inflector) {
        throw new Error("Passing an inflector via PostGraphile options was deprecated in v4.0.0-beta.7; instead please write an inflector plugin: https://www.graphile.org/postgraphile/inflection/");
    }
    const { dynamicJson, classicIds, nodeIdFieldName, replaceAllPlugins, appendPlugins = [], prependPlugins = [], skipPlugins = [], jwtPgTypeIdentifier, jwtSecret, disableDefaultMutations, graphileBuildOptions, graphqlBuildOptions, // DEPRECATED!
    pgColumnFilter, viewUniqueKey, enableTags = true, readCache, writeCache, setWriteCacheCallback, legacyRelations = "deprecated", // TODO:v5: Change to 'omit' in v5
    setofFunctionsContainNulls = true, legacyJsonUuid = false, simpleCollections = "omit", includeExtensionResources = false, ignoreRBAC = true, // TODO:v5: Change to 'false' in v5
    legacyFunctionsOnly = false, // TODO:v5: Remove in v5
    ignoreIndexes = true, } = options;
    if (legacyRelations &&
        ["only", "deprecated", "omit"].indexOf(legacyRelations) < 0) {
        throw new Error("Invalid configuration for legacy relations: " +
            JSON.stringify(legacyRelations));
    }
    if (simpleCollections &&
        ["only", "both", "omit"].indexOf(simpleCollections) < 0) {
        throw new Error("Invalid configuration for simple collections: " +
            JSON.stringify(simpleCollections));
    }
    if (replaceAllPlugins) {
        ensureValidPlugins("replaceAllPlugins", replaceAllPlugins);
        if ((prependPlugins && prependPlugins.length) ||
            (appendPlugins && appendPlugins.length)) {
            throw new Error("When using 'replaceAllPlugins' you must not specify 'appendPlugins'/'prependPlugins'");
        }
    }
    if (readCache && writeCache) {
        throw new Error("Use `readCache` or `writeCache` - not both.");
    }
    let persistentMemoizeWithKey; // NOT null, otherwise it won't default correctly.
    let memoizeCache = {};
    if (readCache) {
        const cacheString = await new Promise((resolve, reject) => {
            fs.readFile(readCache, "utf8", (err, data) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve(data);
                }
            });
        });
        try {
            memoizeCache = JSON.parse(cacheString);
        }
        catch (e) {
            throw new Error(`Failed to parse cache file '${readCache}', perhaps it is corrupted? ${e}`);
        }
    }
    if (readCache || writeCache) {
        persistentMemoizeWithKey = (key, fn) => {
            if (!(key in memoizeCache)) {
                if (readCache) {
                    throw new Error(`Expected cache to contain key: ${key}`);
                }
                memoizeCache[key] = fn();
                if (memoizeCache[key] === undefined) {
                    throw new Error(`Cannot memoize 'undefined' - use 'null' instead`);
                }
            }
            return memoizeCache[key];
        };
    }
    if (writeCache && setWriteCacheCallback) {
        setWriteCacheCallback(() => awaitKeys(memoizeCache).then(obj => new Promise((resolve, reject) => {
            fs.writeFile(writeCache, JSON.stringify(obj), err => {
                memoizeCache = {};
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        })));
    }
    else if (writeCache) {
        throw new Error("Cannot write cache without 'setWriteCacheCallback'");
    }
    else if (setWriteCacheCallback) {
        setWriteCacheCallback(() => Promise.resolve());
    }
    ensureValidPlugins("prependPlugins", prependPlugins);
    ensureValidPlugins("appendPlugins", appendPlugins);
    ensureValidPlugins("skipPlugins", skipPlugins);
    const inflectionOverridePlugins = classicIds
        ? [exports.PostGraphileInflectionPlugin, exports.PostGraphileClassicIdsInflectionPlugin]
        : [exports.PostGraphileInflectionPlugin];
    const basePluginList = replaceAllPlugins
        ? [
            ...prependPlugins,
            ...replaceAllPlugins,
            ...inflectionOverridePlugins,
            ...appendPlugins,
        ]
        : [
            ...prependPlugins,
            ...graphile_build_1.defaultPlugins,
            ...graphile_build_pg_1.defaultPlugins,
            ...inflectionOverridePlugins,
            ...appendPlugins,
        ];
    const invalidSkipPlugins = skipPlugins.filter(pluginToSkip => basePluginList.indexOf(pluginToSkip) < 0);
    if (invalidSkipPlugins.length) {
        function getFunctionName(fn) {
            return fn.displayName || fn.name || String(fn);
        }
        throw new Error(`You tried to skip plugins that would never have been loaded anyway. Perhaps you've made a mistake in your skipPlugins list, or have sourced the plugin from a duplicate plugin module - check for duplicate modules in your 'node_modules' folder. The plugins that you requested to skip were: ${invalidSkipPlugins
            .map(getFunctionName)
            .join(", ")}`);
    }
    const finalPluginList = basePluginList.filter(p => skipPlugins.indexOf(p) === -1);
    return graphile_build_1.getBuilder(finalPluginList, Object.assign({ pgConfig, pgSchemas: Array.isArray(schemas) ? schemas : [schemas], pgExtendedTypes: !!dynamicJson, pgColumnFilter: pgColumnFilter || (() => true), pgInflection: errorProxy("The PostGraphile options `pgInflection` key was deprecated in v4.0.0-beta.7; instead please use `build.inflection` (and note that the call signatures are now simpler): https://www.graphile.org/postgraphile/inflection/"), nodeIdFieldName: nodeIdFieldName || (classicIds ? "id" : "nodeId"), pgJwtTypeIdentifier: jwtPgTypeIdentifier, pgJwtSecret: jwtSecret, pgDisableDefaultMutations: disableDefaultMutations, pgViewUniqueKey: viewUniqueKey, pgEnableTags: enableTags, pgLegacyRelations: legacyRelations, pgLegacyJsonUuid: legacyJsonUuid, persistentMemoizeWithKey, pgForbidSetofFunctionsToReturnNull: !setofFunctionsContainNulls, pgSimpleCollections: simpleCollections, pgIncludeExtensionResources: includeExtensionResources, pgIgnoreRBAC: ignoreRBAC, pgLegacyFunctionsOnly: legacyFunctionsOnly, pgIgnoreIndexes: ignoreIndexes }, graphileBuildOptions, graphqlBuildOptions));
};
function abort(e) {
    /* tslint:disable no-console */
    console.error("Error occured whilst writing cache");
    console.error(e);
    /* tslint:enable no-console */
    process.exit(1);
}
exports.createPostGraphileSchema = async (pgConfig, schemas, options = {}) => {
    let writeCache;
    const builder = await getPostGraphileBuilder(pgConfig, schemas, Object.assign({}, options, { setWriteCacheCallback(fn) {
            writeCache = fn;
        } }));
    const schema = builder.buildSchema();
    if (writeCache) {
        await writeCache().catch(abort);
    }
    return schema;
};
/*
 * Unless an error occurs, `onNewSchema` is guaranteed to be called before this promise resolves
 */
exports.watchPostGraphileSchema = async (pgConfig, schemas, options = {}, onNewSchema) => {
    if (typeof onNewSchema !== "function") {
        throw new Error("You cannot call watchPostGraphileSchema without a function to pass new schemas to");
    }
    if (options.readCache) {
        throw new Error("Using readCache in watch mode does not make sense.");
    }
    let writeCache;
    const builder = await getPostGraphileBuilder(pgConfig, schemas, Object.assign({}, options, { setWriteCacheCallback(fn) {
            writeCache = fn;
        } }));
    let released = false;
    function handleNewSchema(schema) {
        if (writeCache) {
            writeCache().catch(abort);
        }
        onNewSchema(schema);
    }
    await builder.watchSchema(handleNewSchema);
    return async function release() {
        if (released) {
            return;
        }
        released = true;
        await builder.unwatchSchema();
    };
};
// Backwards compat
exports.postGraphQLBaseOverrides = exports.postGraphileBaseOverrides;
exports.postGraphQLClassicIdsOverrides = exports.postGraphileClassicIdsOverrides;
exports.postGraphQLInflection = exports.postGraphileInflection;
exports.postGraphQLClassicIdsInflection = exports.postGraphileClassicIdsInflection;
exports.createPostGraphQLSchema = exports.createPostGraphileSchema;
exports.watchPostGraphQLSchema = exports.watchPostGraphileSchema;
//# sourceMappingURL=index.js.map